/*
 * linux/drivers/pcsnd/pcsp.c
 *
 * /dev/pcsp implementation
 *
 * Copyright (C) 1993-1996  Michael Beck 
 */

#include <linux/config.h>
#include <linux/pcsp.h>

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

/* 16bit emulation needs the mixer */
#ifdef CONFIG_PCSP_16BIT
#define CONFIG_PCSP_MIXER
#endif

extern struct file_operations pcspeaker_pcsp_fops, pcsp_mixer_fops;
extern long pcsp_driver_init(long);
extern int pcsp_set_format(int);
extern void pcsp_free_ports(void);

static int pcsp_open_dev(struct inode * inode, struct file * file)
{
	int minor = MINOR(inode->i_rdev);

	if (! pcsp_enabled)
		return -ENODEV;

	switch (minor) {
		case PCSP_DSP_MINOR:
		case PCSP_AUD_MINOR:
#ifdef CONFIG_PCSP_16BIT
		case PCSP_DSP16_MINOR:
#endif
			file->f_op = &pcspeaker_pcsp_fops;
			break;
#ifdef CONFIG_PCSP_MIXER
		case PCSP_MIXER_MINOR:
			file->f_op = &pcsp_mixer_fops;
			break;
#endif
		default:
			return -ENODEV;
	}
        return file->f_op->open(inode,file);
}

struct file_operations pcsp_fops = {
        NULL,		/* seek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	NULL,		/* ioctl */
	NULL,		/* mmap */
        pcsp_open_dev,
        NULL		/* release */
};

#ifdef MODULE

int init_module(void)
{
	int minor, major, err;

        if (! pcsp_enabled) {
                printk("pcsp disabled\n");
                return -ENODEV;
        }
	
	if ((err = register_chrdev(PCSP_MAJOR, "pcsp", &pcsp_fops))) {
		printk("unable to get major %d for pcsp devices\n", PCSP_MAJOR);
		return err;
	}

	/* remove all symbols */
	register_symtab(0);

	major = (SOUND_VERSION >> 8);
	minor = (SOUND_VERSION & 0xFF);
	if (! (minor & 0xF))
		minor >>= 4;
	printk("pcsp %d.%x:", major, minor);
	pcsp_device_init();
#ifdef CONFIG_PCSP_MIXER
	pcsp_mixer_init();
	printk(", PCSP-mixer");
#endif
	printk(" loaded\n");
	return 0;
}

void cleanup_module(void)
{
	if (MOD_IN_USE)
		printk("pcsnd: busy - remove delayed\n");
	else {
		pcsp_free_ports();
		unregister_chrdev(PCSP_MAJOR, "pcsp");
	}
}
#endif
