/*
 * linux/drivers/pcsnd/pcsp_stub.c
 *
 * /dev/pcsp implementation
 *
 * Copyright (C) 1993-1996  Michael Beck 
 *
 * if PCSP is compiled as a module, this part must
 * be linked with the kernel
 */

#include <linux/config.h>
#include <linux/pcsp.h>

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/timex.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/system.h>

/*
 * need this macros
 */
#define MIN(a,b)        ( ((a) < (b)) ? (a) : (b) )
#define MAX(a,b)        ( ((a) > (b)) ? (a) : (b) )

/*
 * need this to be global
 */
char pcsp_active = 0;
char pcsp_speaker = 0;
char pcsp_enabled = -1;
volatile int pcsp_timer0_latch, pcsp_clockticks;
volatile int pcsp_test_running;

struct pcsp_status pcsp;

extern struct file_operations pcsp_fops;


#ifndef CONFIG_PCSP_NO_TEST_SPEED

static void *sleep = NULL;

/*
   this is a stupid beep which occurs if PCSP is disabled;
   it's not needed because we have the message, but who reads it...
   and this is the PC-Speaker driver :-)
*/
void pcsp_beep(int count, int cycles)
{
	/* enable counter 2 */
	outb_p(inb_p(0x61)|3, 0x61);
	/* set command for counter 2, 2 byte write */
	outb_p(0xB6, 0x43);
	/* select desired HZ */
	outb_p(count & 0xff, 0x42);
	outb((count >> 8) & 0xff, 0x42);

	while (cycles--);
		 
	/* disable counter 2 */
	outb(inb_p(0x61)&0xFC, 0x61);
}

/*
   the timer-int for testing cpu-speed, mostly the same as
   for PC-Speaker
 */
static int pcsp_test_intspeed(void)
{
	if (pcsp.index < pcsp.in[pcsp.actual]) {
		outb(pcsp.e, 0x61);
		__asm__("xorb $0,%0\n\t"
			"outb %0,$97"
			: : "a" ((char)pcsp.e) );
		__asm__("xlatb\n\t"
			"outb %0,$66"
			: : "a" ((char)pcsp.buffer[pcsp.index]),
			    "b" ((long)pcsp.buffer) );
		pcsp.xfer += pcsp.si;
		pcsp.index = pcsp.xfer >> 16;
	}
	if (pcsp.index >= pcsp.in[pcsp.actual]) {
		pcsp.xfer = pcsp.index = 0;
		pcsp.in[pcsp.actual] = 0;
		pcsp.actual ^= 1;
		pcsp.buffer = pcsp.buf[pcsp.actual];
                if (sleep)	/* NEVER */
			nop();
		if (pcsp.in[pcsp.actual] == 0xFFFF)
			pcsp.actual ^= 1;
	}

	if ( (pcsp_clockticks -= pcsp.timerCF) < 0)
		pcsp_clockticks += LATCH;
	++pcsp_test_running;
	return 1;
}

/*
   this routine measures the time needed for one timer-int if
   we play thru PC-Speaker. This is kind of ugly but does the
   trick.
 */
static int pcsp_measurement(unsigned char *buf, int addon)
{
	int count;
	unsigned long flags;

	pcsp_clockticks   = 0;
	pcsp.timerCF	  = LATCH;
	pcsp.buf[0]       =
	pcsp.buffer       = buf;
	pcsp.index	  = 0;
	pcsp.xfer	  = 0;
	pcsp.si		  = 1 << 16;
	pcsp.in[0]        = 5 + addon;
	pcsp.in[1] 	  = 0;
	pcsp.actual       = 0;
	pcsp.e 		  = inb(0x61) & 0xFC;

	pcsp_test_running = 0;

	if (pcsp_set_irq(pcsp_test_intspeed) < 0)
		panic("PCSP could not modify timer IRQ!");

	/*
	   Currently (0.99.15d) Linux call chr_dev_init with ints
	   disabled; so we need a sti() to enable them.
	   However, because this can be changed in the future we use
	   save_flags() and restore_flags()
	*/
	save_flags(flags);
	sti();

	/*
	  Perhaps we need some sort of timeout here, but if IRQ0
	  isn't working the system hangs later ...
	*/
	while (pcsp_test_running < 5);
	restore_flags(flags);

	if (pcsp_release_irq() < 0)
		panic("PCSP could not reset timer IRQ!");

	outb_p(0x00, 0x43);		/* latch the count ASAP */
	count = inb_p(0x40);		/* read the latched count */
	count |= inb(0x40) << 8;
	return (LATCH - count);
}

static int pcsp_test_speed(void)
{
	int worst, worst1, best, best1;
	unsigned char test_buffer[32];

	worst  = pcsp_measurement(test_buffer, 0);
	worst1 = pcsp_measurement(test_buffer, 0);
	best   = pcsp_measurement(test_buffer, 5);
	best1  = pcsp_measurement(test_buffer, 5);

	worst = MAX(worst, worst1);
	best  = MIN(best, best1);

#ifdef PCSP_DEBUG
	printk("  PCSP-Timerint needs %d Ticks in worst case\n", worst);
	printk("  PCSP-Timerint needs %d Ticks in best case\n", best);
#endif
	/* We allow a CPU-usage of 90 % for the best-case ! */
	pcsp.realrate = best * 10 / 9;
	pcsp.maxrate  = CLOCK_TICK_RATE / pcsp.realrate;
	printk(" maximal samplerate %d Hz", pcsp.maxrate);

	if (pcsp.maxrate > PCSP_CRITICAL_FREQ) {
		if (MIN_CONST > pcsp.realrate) {
			pcsp.realrate = MIN_CONST;
			printk(", %d Hz", MAX_SRATE);
		}
		printk(" used\n");
		return 1;
	}

	printk("\n  This is too SLOW! PCSP-driver DISABLED\n");

	/* very ugly beep, but you hopefully never hear it */
	pcsp_beep(12000,800000);
	pcsp_beep(10000,800000);
	
	return 0;
}
#endif

void pcsp_setup(char *s, int *p)
{
	if (!strcmp(s, "off")) {
		pcsp_enabled = 0;
		return;
	}
	if (p[0] > 0 && p[1] > 0)
		pcsp.maxrate = p[1];
	pcsp_enabled = 1;
}

void pcsp_driver_init(void)
{ 
	if (pcsp_enabled < 0) {
#ifndef CONFIG_PCSP_NO_TEST_SPEED
		pcsp_enabled = pcsp_test_speed();
#else
		pcsp.maxrate  = 44100;	/* only a BIG freq */
		pcsp.realrate = (CLOCK_TICK_RATE + CONFIG_PCSP_SRATE / 2) /
				CONFIG_PCSP_SRATE;
		pcsp_enabled  = 1;
#endif
	}
	else {
		pcsp.realrate = MIN(pcsp.maxrate, MAX_SRATE);
		pcsp.realrate = (CLOCK_TICK_RATE + pcsp.realrate / 2) /
				pcsp.realrate;
	}
}

extern int pcsp_set_irq(int (*func)(void));
extern int pcsp_release_irq(void);

static struct symbol_table pcsp_syms = {
#include <linux/symtab_begin.h>
	X(pcsp_timer0_latch),
	X(pcsp_clockticks),
#ifdef CONFIG_PCSP_MODULE
	X(pcsp_enabled),
	X(pcsp_active),
	X(pcsp_speaker),
	X(pcsp),
	X(pcsp_set_irq),
	X(pcsp_release_irq),
#endif
#include <linux/symtab_end.h>
};

/*
 * If your kernel is less than 1.3.33, pcsp must take and return the
 * kmem_start argument that used to be passed to initializing drivers.
 */
#if LINUX_VERSION < (1*65536 + 3*256 + 33*1)
#define init_t long
#define init_args long kmem_start
#define init_retval kmem_start
#else
#define init_t int
#define init_args void
#define init_retval 0
#endif

/*
 * the pcsp_init() function is called in chrdev_init at kernel startup
 */
#ifndef CONFIG_PCSP_MODULE
init_t pcsp_init(init_args)
{
	int minor, major;

	/* if disabled in commandline */
	if (! pcsp_enabled) {
		printk("PCSP-device disabled\n");
		return init_retval;
	}

	if (register_chrdev(PCSP_MAJOR, "pcsp", &pcsp_fops)) {
		printk("unable to get major %d for PCSP-devices\n", PCSP_MAJOR);
		return init_retval;
	}
	major = (SOUND_VERSION >> 8);
	minor = (SOUND_VERSION & 0xFF);
	if (! (minor & 0xF))
		minor >>= 4;
	printk("PCSP %d.%x measurement:", major, minor);
	pcsp_driver_init();
	if (pcsp_enabled) {
		printk("PCSP %d.%x:", major, minor);
		pcsp_device_init();
#if defined(CONFIG_PCSP_16BIT) || defined(CONFIG_PCSP_MIXER)
		pcsp_mixer_init();
		printk(", PCSP-mixer");
#endif
		printk(" installed\n");
	}
	register_symtab(&pcsp_syms);
	return init_retval;
}

#else

/*
 * this pcsp_init() function is called when the driver is loaded as a module
 * at kernel-startup
 */
init_t pcsp_init(init_args)
{
	int minor, major;

	major = (SOUND_VERSION >> 8);
	minor = (SOUND_VERSION & 0xFF);
	if (! (minor & 0xF))
		minor >>= 4;
	printk("PCSP %d.%x measurement:", major, minor);
	pcsp_driver_init();
	register_symtab(&pcsp_syms);
	return init_retval;
}
#endif
